/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:10:56 1991 by burrows        */
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include "indexserver.h"

static struct msgset *concord1 ();
static struct msgset *find ();
static struct msgset *ffind ();
static int makeprefix ();

extern int prefixmatch ();

struct msgset *concord (index2, msgset, origpat, ignorecase, matchtype)
struct index2 *index2;
struct msgset *msgset;
char *origpat;
int ignorecase;
int matchtype;
{
	int (*match) () = prefixmatch;
	char patprefix[128];
	int i = 0;
	char *pat = origpat;
	char *shellpat = 0;
	char var[128];
	int found = 0;
	struct msgset *varmsgset;

	if (matchtype == SHELLMATCH) {
		match = notfindsimpleshellmeta;
		while (origpat[i] == '*')
			i++;
	} else if (matchtype == EXACTMATCH)
		match = strcmp;

	if (origpat[i] == '\0') {
		freemsgset (msgset);
		return (ormsgset (index2->allmsgs, (struct msgset *) 0));
	}

	var[0] = '\0';
	if (strlen (pat) < sizeof (var) - 20) {
		(void) sprintf (var, "-%d,%d,%s", ignorecase?1:0, matchtype,
			pat);
		varmsgset = getvar (var, &found);
		if (found) {
/* (void) fprintf (stderr, "concord cache hit on %s\n", var); */
			return (into_ormsgset (msgset, varmsgset));
		}
/* (void) fprintf (stderr, "concord cache miss on %s\n", var); */
	}

	if (ignorecase) {
		pat = cipat (pat, match == notfindsimpleshellmeta);
		if (match == prefixmatch)
			(void) strcat (pat, "*");
		match = notfindsimpleshellmeta;
	}

	if (match == notfindsimpleshellmeta)
		shellpat = compileshellmeta (pat, 1);

	i = 0;
	do {
		i = makeprefix (pat, shellpat != 0, 
			    i, patprefix, sizeof (patprefix));

		if (index2->indices[0].concord.data.ndata > 0)
			msgset = concord1 (&index2->indices[0], msgset,
				match, patprefix, pat, shellpat);
		if (index2->indices[1].concord.data.ndata > 0)
			msgset = concord1 (&index2->indices[1], msgset,
				match, patprefix, pat, shellpat);
	} while (i > 0);

	if (pat != origpat)
		free (pat);
	if (shellpat != 0)
		freeshellmeta (shellpat);
	
	if (var[0] != '\0')
		setvar (var, msgset);
	
	return (msgset);
}

static struct msgset *concord1 (index, msgset, match, patprefix, pat, re)
struct index *index;
struct msgset *msgset;
int (*match) ();
char *patprefix;
char *pat;
char *re;
{
	long lastpos = 0;
	char buf[2000];
	int l = strlen (patprefix);
	register char *data = index->concord.subindex.text.data;
	register long *lines = index->concord.subindex.lines;
	long nlines = index->concord.subindex.nlines;
	long pos;

	if (l > 0) {
		long lo = 0, hi = nlines, mid;

		while (lo < hi-1) {
			mid = (lo+hi) >> 1;

			if (strncmp (data+lines[mid], patprefix, l) >= 0)
				hi = mid;
			else
				lo = mid;
		}

		while (lo < nlines) {
			if (strncmp (data+lines[lo], patprefix, l) >= 0)
				break;
			lastpos = lo;
			lo++;
		}
	}

	if (lastpos < nlines) {
		(void) strcpy (buf, data+lines[lastpos]);
		pos = atol (data+lines[lastpos]+strlen (buf)+1);
	} else {
		(void) strcpy (buf, "");
		pos = 0;
	}

	return (ffind (index, pos, buf, (re?patprefix:""),
		(re?re:pat), match, msgset));
}

static struct msgset *find (index, pos, buf, patprefix, pat, match, msgset)
struct index *index;
long pos;
char *buf;
char *patprefix;
char *pat;
int (*match) ();
struct msgset *msgset;
{
        register int c;
	register char *p, *e;
        register char *s;
        register long n, ln = 0;
	int cmp;
	int l = strlen (patprefix);
	int shift;
	char *data;
	long *bitset;
	long maxmsg;

	if (msgset == 0) {
		maxmsg = 0;
		bitset = 0;
	} else {
		maxmsg = msgset->nbitset * LONGBITS;
		bitset = msgset->bitset;
	}

	data = index->concord.data.data;
	p = data + pos;
	e = data + index->concord.data.ndata;

	*e++ = 0x80;
	c = *p++ & 0xff;

        while ((c & 0xc0) == 0x80 && p < e && !interrupt) {
                c &= 0x3f;
                s = &buf[c];

		e[-1] = '\0';
                while ((*s++ = *p++) != '\0' && (*s++ = *p++) != '\0' &&
                       (*s++ = *p++) != '\0' && (*s++ = *p++) != '\0')
                        ;
		e[-1] = 0x80;

		if (l == 0) {
			cmp = (*match) (pat, buf);
			if (cmp < 0)
				break;
		} else {
			cmp = strncmp (patprefix, buf, l);

			if (cmp < 0)
				break;
			if (cmp == 0)
				cmp = (*match) (pat, buf);
		}

		if (cmp == 0) {
			c = *p++ & 0xff;
			ln = 0;
			while ((c & 0xc0) != 0x80) {
				n = 0;
				shift = 0;
				while ((c & 0xc0) == 0xc0) {
					n |= (c & 0x3f) << shift;
					shift += 6;
					c = *p++ & 0xff;
				}
				n |= (c & 0x7f) << shift;
				n += ln;
				ln = n;
				if (n >= maxmsg) {
					msgset = extendmsgset (msgset, 
						(int) ROUNDUP (n+200000));
					maxmsg = msgset->nbitset * LONGBITS;
					bitset = msgset->bitset;
				}
				bitset[ROUNDDOWN (n)] |= 1 << BIT (n);
				c = *p++ & 0xff;
			}
		} else {
			while ((*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80 &&
			       (*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80)
				;
			c = p[-1] & 0xff;
		}
        }

	return (msgset);
}

static char *fill (fd, p, e, data, readsize)
int fd;
char *p;
char *e;
char *data;
int readsize;
{
	int n;

	if (p > e+1) {
		(void) fprintf (stderr, "indexserver: fill underrun\n");
		abort ();
	}
	if (p > e)
		p = e;
	if (e > p)
		bcopy (p, data, e-p);
	n = read (fd, data + (e-p), readsize);
	if (n < 0) {
		perror ("indexserver: fill: read");
		exit (4);
	}
	return (data + (e - p) + n);
}

#define ENSIZE		10
#define READSIZE	8192

static struct msgset *ffind (index, pos, buf, patprefix, pat, match, msgset)
struct index *index;
long pos;
char *buf;
char *patprefix;
char *pat;
int (*match) ();
struct msgset *msgset;
{
        register int c;
	register char *p, *e;
        register char *s;
        register long n, ln = 0;
	int cmp;
	int l = strlen (patprefix);
	int shift;
	char *data;
	long *bitset;
	long maxmsg;
	char stackdata[READSIZE+ENSIZE+1];
	int fd = -1;

	if (msgset == 0) {
		maxmsg = 0;
		bitset = 0;
	} else {
		maxmsg = msgset->nbitset * LONGBITS;
		bitset = msgset->bitset;
	}

#define ENSURE()	\
	((p+ENSIZE >= e && fd >= 0)?((e = fill (fd, p, e-1, data, READSIZE)), \
			(p = data), (*e++ = 0x80),1):1)

	data = index->concord.data.data;
	if (data == 0) {
		p = e = data = stackdata;
		*e++ = 0x80;
		fd = index->concord.fd;
		if (lseek (fd, pos, 0) == -1L) {
			perror ("indexserver: ffind: lseek");
			exit (2);
		}
		ENSURE ();
	} else {
		p = data + pos;
		e = data + index->concord.data.ndata;

		*e++ = 0x80;
	}

	c = *p++ & 0xff;

        while ((c & 0xc0) == 0x80 && p < e) {
                c &= 0x3f;
                s = &buf[c];

		e[-1] = '\0';
                while (ENSURE () &&
		       (*s++ = *p++) != '\0' && (*s++ = *p++) != '\0' &&
		       (*s++ = *p++) != '\0' && (*s++ = *p++) != '\0' &&
		       (*s++ = *p++) != '\0' && (*s++ = *p++) != '\0' &&
		       (*s++ = *p++) != '\0' && (*s++ = *p++) != '\0')
                        ;
		e[-1] = 0x80;

		if (l == 0) {
			cmp = (*match) (pat, buf);
			if (cmp < 0)
				break;
		} else {
			cmp = strncmp (patprefix, buf, l);

			if (cmp < 0)
				break;
			if (cmp == 0)
				cmp = (*match) (pat, buf);
		}

		if (cmp == 0) {
			c = *p++ & 0xff;
			ln = 0;
			while ((c & 0xc0) != 0x80) {
				n = 0;
				shift = 0;
				if ((c & 0xc0) == 0xc0) {
					ENSURE ();
				}
				while ((c & 0xc0) == 0xc0) {
					n |= (c & 0x3f) << shift;
					shift += 6;
					c = *p++ & 0xff;
				}
				n |= (c & 0x7f) << shift;
				n += ln;
				ln = n;
				if (n >= maxmsg) {
					msgset = extendmsgset (msgset, 
						(int) ROUNDUP (n+40000));
					maxmsg = msgset->nbitset * LONGBITS;
					bitset = msgset->bitset;
				}
				bitset[ROUNDDOWN (n)] |= 1 << BIT (n);
				ENSURE ();
				c = *p++ & 0xff;
			}
		} else {
			while (ENSURE () &&
			       (*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80 &&
			       (*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80 &&
			       (*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80 &&
			       (*p++ & 0xc0) != 0x80 && (*p++ & 0xc0) != 0x80)
				;
			c = p[-1] & 0xff;
		}
        }

	return (msgset);
}

int prefixmatch (pat, buf)
register char *pat;
register char *buf;
{
	while (*pat == *buf && *buf != '\0')
		pat++, buf++;
	
	if (*pat == '\0')
		return (0);
	return ((*pat & 0xff) - (*buf & 0xff));
}

/* case independent pattern  from pattern or string */
char *cipat (s, pat)
register char *s;
int pat;
{
	char *ipat = malloc ((unsigned)((strlen (s)+1) * 4));
	register char *p = ipat;

	if (ipat == 0) {
		perror ("malloc");
		exit (2);
	}

	while (*s != '\0') {
		if (pat && *s == '[' && s[1] != '\0' && s[2] != '\0' &&
				index (s+(s[1]=='!'?3:2), ']') != 0) {
                        *p++ = *s++;
                        if (*s == '!')
                                *p++ = *s++;
			do {
                                if (s[1]=='-'&& s[2] != '\0' && s[2] != ']') {
                                        register int i, j, lo = 0xff, hi = 0;

                                        i = tolower (s[0] & 0xff);
                                        j = tolower (s[2] & 0xff);
                                        while (i <= j) {
                                                if (isascii (i) && isalpha (i)){
                                                        if (i < lo)
                                                                lo = i;
                                                        if (i > hi)
                                                                hi = i;
                                                }
                                                i++;
                                        }
                                        if (lo <= hi) {
                                                *p++ = toupper (lo);
                                                if (lo < hi) {
                                                        *p++ = '-';
                                                        *p++ = toupper (hi);
                                                }
                                        }
                                        *p++ = tolower (s[0] & 0xff);
                                        *p++ = '-';
					s += 2;
                                } else if (isascii (*s) && isalpha (*s))
                                        *p++ = toupper (*s);
                                *p++ = tolower (*s);
				s++;
                        } while (*s != '\0' && *s != ']');
		} else if (isascii (*s) && isalpha (*s)) {
			*p++ = '[';
			*p++ = toupper (*s);
			*p++ = tolower (*s);
			s++;
			*p++ = ']';
		} else if (!pat && (*s=='[' || *s=='?' || *s == '*')) {
			*p++ = '[';
			*p++ = *s++;
			*p++ = ']';
		} else
			*p++ = *s++;
	}
	*p = '\0';
	return (ipat);
}


static int makeprefix (pat, re, count, prefix, size)
char *pat;
int re;
int count;
char *prefix;
int size;
{
	register int i, j, k;
	register char *p, *s;

	if (!re) {
		(void) strncpy (prefix, pat, size);
		prefix[size-1] = '\0';
		return (0);
	}

	i = 0;
	while (i < size-1 && pat[i] != '?' &&
		    pat[i] != '*' && pat[i] != '[' && pat[i] != '\0') {
		prefix[i] = pat[i];
		i++;
	}

	prefix[i] = '\0';
	if (i > 3 || i >= size-3 || pat[i] != '[')
		return (0);

	p = &pat[i+1];
	for (j=0; j<5 && p[j]!=']' && p[j]!='!' && p[j]!='-' && p[j]!='\0'; j++)
		;
	if (j == 0 || j >= 5 || p[j] != ']')
		return (0);

	if (j < 3 && p[j+1] == '[') {
		s = &p[j+2];
		for (k=0; k<3 && s[k]!=']' && s[k]!='!' &&
			  s[k]!='-' && s[k]!='\0'; k++)
			;
		if (0 < k && k < 3 && s[k] == ']') {
			if (count >= j*k)
				count = j*k - 1;
			prefix[i++] = p[count / k];
			prefix[i++] = s[count % k];
			prefix[i] = '\0';
			if (count+1 < j*k)
				return (count+1);
			return (0);
		}
	}
	
	if (count >= j)
		count = j-1;
	prefix[i++] = p[count];
	p = &p[j+1];
	while (i < size-1 && *p != '\0' && *p != '?' && *p != '*' && *p != '[')
		prefix[i++] = *p++;
	prefix[i] = '\0';
	if (count+1 < j)
		return (count+1);
	return (0);
}
