/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Feb 24 16:11:02 1991 by burrows        */
struct binfile {
	long ndata;		/* bytes of data */
	char *data;		/* malloced data */
};

struct textfile {
	struct binfile text;	/* data */
	long nlines;		/* number of lines */
	long *lines;		/* malloced array of offsets of each line */
};

struct concord {
	struct binfile data;		/* data */
	int fd;				/* if data.data == 0, file desc */
	struct textfile subindex;	/* subindex for concordance */
					/* each line is the same as one entry */
					/* in the data. After the terminating */
					/* null at the end of the line, there */
					/* is an ascii string giving the */
					/* offset of the entry in data, */
					/* in decimal */
};

struct message {
	unsigned long group;		/* group number */
	unsigned long message;		/* message number */
};

struct groups {
	struct textfile text;		/* groups file */
	long ngroup;			/* number of groups */
	long *group;			/* malloced array */
					/* name of group N is \0 terminated */
					/* string starting at */
					/* text.text.data+group[N] */
};

struct summary {
	struct textfile summary;
	char *fields;
	struct binfile messages;
	struct binfile dates;
};

struct index {
	struct concord concord;		/* concordance for this index */
	struct summary summary;		/* summary for this index */
};

struct index2 {				/* two indices */
	struct index indices[2];
	struct msgset *allmsgs;		/* all messages */
};

struct msgset {
	int nbitset;
	long *bitset;
};

struct messagepat {
	unsigned long message;
	char *grouplist;
};

extern void freeindex ();
extern int readindex ();
extern void freegroups ();
extern int readgroups ();
extern int readtextfile ();
extern int readbinfile ();
extern void freebinfile ();
extern void freetextfile ();

extern void setvar ();
extern void clearvars ();
extern struct msgset *getvar ();

extern struct msgset *getmsgset ();
extern struct msgset *submsgset ();
extern int nc ();

extern struct msgset *notmsgset ();
extern struct msgset *andmsgset ();
extern struct msgset *ormsgset ();
extern struct msgset *into_notmsgset ();
extern struct msgset *into_andmsgset ();
extern struct msgset *into_ormsgset ();
extern struct msgset *newmsgset ();
extern struct msgset *extendmsgset ();
extern void freemsgset ();
extern long countmsgset ();
extern int applymsgset ();

extern struct msgset *concord ();
extern char *cipat ();

extern struct msgset *discussfunc ();

extern struct msgset *dateset ();

extern void tableinitialize ();
extern struct msgset *filterselectmsg ();
extern struct msgset *selectmsg ();
extern struct msgset *restrict ();
extern char *field ();
extern int nullselect ();
extern int onselect ();
extern int afterselect ();
extern int beforeselect ();
extern int fromselect ();
extern int toselect ();
extern int groupselect ();
extern int msgselect ();
extern int ccselect ();
extern int subjselect ();
extern int samesubjselect ();
extern int fromtoccselect ();
extern int toccselect ();
extern int repselect ();
extern int refselect ();
extern int idselect ();
extern int midselect ();
extern int keyselect ();

extern void freeshellmeta ();
extern int findshellmeta ();
extern int notfindsimpleshellmeta ();
extern int findstartshellmeta ();
extern char *compileshellmeta ();

extern int parsedate ();
extern char *unparsedate ();

extern long atol ();
extern long strtol ();
extern void perror ();
extern void exit ();
extern void free ();
extern char *mymalloc ();
extern char *myrealloc ();
extern void bcopy ();
extern long lseek ();
extern long time ();

extern int interrupt;

#define MIN(aa,bb)      ((aa) < (bb)? (aa): (bb))
#define MAX(aa,bb)      ((aa) > (bb)? (aa): (bb))

#define ROUNDDOWN(xx)   ((xx) >> 5)
#define ROUNDUP(xx)     (((xx)+31) >> 5)
#define BIT(xx)         ((xx) & 0x1f)
#define LONGBITS        32

#define ELEMENTS(x)     (sizeof (x) / sizeof ((x)[0]))

#define NFIELDS		10

#define SUBJECTFIELD    0
#define FROMFIELD       1
#define TOFIELD         2
#define CCFIELD         3
#define DATEFIELD       4
#define NEWSGROUPSFIELD 5
#define MESSAGEIDFIELD  6
#define INREPLYTOFIELD  7
#define REFERENCESFIELD 8
#define KEYWORDSFIELD   9

#define PREFIXMATCH	0
#define EXACTMATCH	1
#define SHELLMATCH	2

/* allow core cumps when running suid */
#define abort		myabort
extern void abort ();

#define malloc(x)	mymalloc(x, __FILE__, __LINE__)
#define realloc(x,y)	myrealloc(x, y, __FILE__, __LINE__)
