/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Sun Mar  3 20:10:17 1991 by burrows        */
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include "indexserver.h"

static char lower[256];
static char notwordstart[256];
static char notinword[256];
static char all[256];

void tableinitialize ()
{
	register int i;

	for (i = 0; i < 256; i++) {
		lower[i] = tolower (i);
		notwordstart[i] = !isalnum (i);
		notinword[i] = !isalnum (i);
		all[i] = 1;
	}
}

static int shellmetalen (p)
register char *p;
{
	register char *s;

	if (*p == '\0')
		return (0);

	if (*p == '[') {
		s = p+1;

		if (*s == '!')
			s++;

		if (*s == ']')
			return (1);

		while (*s != '\0' && *s != ']')
			s++;
		if (*s == ']')
			return (s+1 - p);
	}

	return (1);
}

struct msgset *filterselectmsg (index2, set, func, string, not, literal)
struct index2 *index2;
struct msgset *set;
int (*func) ();
char *string;
int not;
int literal;
{
	struct msgset *result;
	char strbuf[512];
	char *str = strbuf;
	int meta = 0;
	char *pat = 0;
	register long *bitset;
	register long l;
	register int i, j, k;
	struct msgset *tmp;
	int c;


	if (set == 0)
		return (0);
	
	if (*string == '\0')
		return (ormsgset (set, (struct msgset *) 0));

	k = 1;
	for (i=0,j=0; (c=(string[i]&0xff)) != '\0' && j<sizeof (strbuf)-1; i++) {
		if (isspace (c)) {
			if (!k)
				str[j++] = c;
			k = 1;
		} else {
			meta = meta || c == '*' || c == '?' || c == '[';
			str[j++] = lower [c];
			k = 0;
		}
	}
	str[j] = '\0';

	if (literal)
		meta = 0;

	result = restrict (index2, set, str, meta);

	if (literal)
		str = string;

	if (meta) {
		/* str = cipat (str, meta); */
		pat = compileshellmeta (str, 1);
	}

	if (result != 0) {
		bitset = result->bitset;
		for (i = 0, j = result->nbitset; i < j; i++) {
			if ((l = bitset[i]) != 0) {
				for (k = 0; k < LONGBITS; k++) {
					if (l & (1 << k)) {
						if (!(*func) (index2,
						    k+i*LONGBITS, str, pat))
							l &= ~(1 << k);
					}
				}
				bitset[i] = l;
			}
		}
	}

	if (not) {
		tmp = notmsgset (set, result);
		freemsgset (result);
		result = tmp;
	}

	/* if (meta && str != strbuf) free (str); */

	if (pat != 0)
		freeshellmeta (pat);

	return (result);
}

struct msgset *restrict (index2, msgset, string, meta)
struct index2 *index2;
struct msgset *msgset;
char *string;
int meta;
{
	char buf[129];
	register char *p = string, *q;
	register char *nws = notwordstart;
	register char *niw = notinword;
	struct msgset *tmp, *result;
	int ml;

	result = ormsgset (msgset, (struct msgset *) 0);

	while (*p != 0 && countmsgset (result) > 4000) {
		if (meta) {
			for (;;) {
				while (*p!='\0' && *p != '[' && nws[*p & 0xff])
					p++;

				if (*p == '[')
					p += shellmetalen (p);
				
				if (*p == '\0')
					return (result);
				if (p == string || index ("*?]", p[-1]) == 0)
					break;

				for (;;) {
					if (!niw[*p & 0xff] || *p=='*' || *p=='?')
						p++;
					else if (*p == '[' && 
					    (ml=shellmetalen (p)) > 1)
						p += ml;
					else
						break;
				}
			}
			
			q = buf;
			for (;;) {
				if (!niw[*p & 0xff]) {
					if (q < &buf[sizeof (buf)-1])
						*q++ = *p;
					p++;
				} else if (*p == '*' || *p == '?' || *p=='[') {
					/* can't continue * and ? match space */
					if (q < &buf[sizeof (buf)-1])
						*q++ = '*';
					if (q < &buf[sizeof (buf)-1])
						*q++ = '\0';

					if (*p=='[' && (ml=shellmetalen(p)) > 1)
						p += ml;
					else
						p++;
				} else
					break;
			}
			*q = '\0';
		} else {
			while (*p != '\0' && nws[*p & 0xff])
				p++;
			
			if (*p == '\0')
				return(result);

			q = buf;
			while (!niw[*p & 0xff]) {
				if (q < &buf[sizeof (buf)-1])
					*q++ = *p;
				p++;
			}
			*q = '\0';
		}

		tmp = concord (index2, (struct msgset *)0, buf,
			0, meta?SHELLMATCH:EXACTMATCH);
		result = into_andmsgset (result, tmp);
	
		freemsgset (tmp);
/* (void) fprintf (stderr, "restrict %s -> %d\n", buf, countmsgset (result)); */
	}

	return (result);
}

struct msgset *selectmsg (index2, set, func, string, not)
struct index2 *index2;
struct msgset *set;
int (*func) ();
char *string;
int not;
{
	struct msgset *result;
	register int ok = (not?0:1);
	register long *bitset;
	register long l;
	register int i, j, k;

	if (set == 0)
		return (0);

	result = newmsgset (set->nbitset);

	bcopy ((char *) set->bitset, (char *) result->bitset,
		result->nbitset * sizeof (long));

	bitset = result->bitset;
	for (i = 0, j = result->nbitset; i < j; i++) {
		if ((l = bitset[i]) != 0) {
			for (k = 0; k < LONGBITS; k++) {
				if (l & (1 << k)) {
					if (((*func) (index2,
						(long) k+i*LONGBITS,
						string, (char *) 0)?1:0) != ok)
						l &= ~(1 << k);
				}
			}
			bitset[i] = l;
		}
	}

	return (result);
}

char *field (index2, msg, fieldnum)
struct index2 *index2;
long msg;
register int fieldnum;
{
	register int c;
	register struct index *index;
	register char *p;
	register long n;

	if (msg < (n = index2->indices[0].summary.summary.nlines)) {
		index = &index2->indices[0];
	} else if (msg < n+index2->indices[1].summary.summary.nlines) {
		index = &index2->indices[1];
		msg -= n;
	} else {
		(void) fprintf (stderr,
		 "indexserver: field: message %d not in either index (%d %d)\n",
			msg, index2->indices[0].summary.summary.nlines,
			index2->indices[1].summary.summary.nlines);
		abort ();
	}

	if (fieldnum < 0 || fieldnum >= NFIELDS) {
		(void) fprintf (stderr,
		    "indexserver: field: bad field number %d\n", fieldnum);
		abort ();
	}

	c = index->summary.fields[msg * 10 + fieldnum] & 0xff;

	p = index->summary.summary.text.data +
	    index->summary.summary.lines[msg];

	if (c > 0)
		return (p + c);

	fieldnum++;
	do {
		while (*p++ != '\0')
			;
	} while (--fieldnum != 0);

	return (p);
}

/* ARGSUSED */
int search (field, string, pat)
char *field;
char *string;
char *pat;
{
	register char *p = field, *q, *r;
	register char *l = lower;
	register char *niw = notinword;

	if (pat != 0) {
		if (findshellmeta (pat, field, notinword))
			return (1);
		if (*string == '*')
			return (0);
		return (findstartshellmeta (pat, field,notwordstart,notinword));
	} else {
		int c = *string;
		int w = !notwordstart[c & 0xff];

		while (*p != '\0') {
			if (w) {
				while (*p != '\0' && l[*p & 0xff] != c)
					p++;
				if (*p == '\0')
					return (0);
				while (p > field && !niw[p[-1] & 0xff]) {
					do {
						p++;
					} while (*p!='\0' && l[*p & 0xff] != c);
					if (*p == '\0')
						return (0);
				}
				r = p++;
			} else {
				while (*p != '\0' && l[*p & 0xff] != c)
					p++;
				if (*p == '\0')
					return (0);
				r = p++;
			}

			q = string;
			while (*r != '\0' && *q == l[*r & 0xff])
				r++, q++;
			
			if (*q == '\0') {
				if (niw[q[-1] & 0xff] || niw[*r & 0xff])
					return (1);
			}
		}
	}

	return (0);
}

/* ARGSUSED */
int nullselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
        return (0);
}

/* ARGSUSED */
int onselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
register char *string;
char *pat;
{
	register char *p = field (index2, msg, DATEFIELD);

	while (*p == *string && *p != '\0')
		p++, string++;

	return (*string == '\0');
}

/* ARGSUSED */
int afterselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
register char *string;
char *pat;
{
	register char *p = field (index2, msg, DATEFIELD);

	while (*p == *string && *p != '\0')
		p++, string++;

	return (*string != '\0' && *p > *string);
}

/* ARGSUSED */
int beforeselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
register char *string;
char *pat;
{
	register char *p = field (index2, msg, DATEFIELD);

	while (*p == *string && *p != '\0')
		p++, string++;

	return (*string != '\0' && *p < *string);
}

int fromtoccselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, FROMFIELD), string, pat) ||
	        search (field (index2, msg, TOFIELD), string, pat) || 
	        search (field (index2, msg, CCFIELD), string, pat));
}

int fromselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, FROMFIELD), string, pat));
}

int toselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, TOFIELD), string, pat));
}

int toccselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, TOFIELD), string, pat) || 
	        search (field (index2, msg, CCFIELD), string, pat));
}

/* ARGSUSED */
int msgselect (index2, msg, mpat, pat)
register struct index2 *index2;
long msg;
register struct messagepat *mpat;
char *pat;
{
	register struct message *m;
	register long mindex = msg * sizeof (struct message);
	register long n;

        if (mindex < (n = index2->indices[0].summary.messages.ndata)) {
                m = (struct message *)
                        (index2->indices[0].summary.messages.data + mindex);
        } else if (mindex < n + index2->indices[1].summary.messages.ndata) {
                m = (struct message *)
                        (index2->indices[1].summary.messages.data + mindex - n);
        } else {
		(void) fprintf (stderr,
		    "indexserver: msgselect: message %d not in either index (%d %d %d)\n",
			msg, mindex, index2->indices[0].summary.messages.ndata,
			index2->indices[1].summary.messages.ndata);
                abort ();
	}

	if (mpat->message != 0 && mpat->message != m->message)
		return (0);

	return (mpat->grouplist[m->group]);
}

int groupselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, NEWSGROUPSFIELD), string, pat));
}

int keyselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, KEYWORDSFIELD), string, pat));
}

int idselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, MESSAGEIDFIELD), string, pat) ||
		search (field (index2, msg, INREPLYTOFIELD), string, pat));
		
}

int midselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, MESSAGEIDFIELD), string, pat));
}

int repselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, INREPLYTOFIELD), string, pat));
}

int refselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, REFERENCESFIELD), string, pat));
}

int ccselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, CCFIELD), string, pat));
}

int subjselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
char *string;
char *pat;
{
	return (search (field (index2, msg, SUBJECTFIELD), string, pat));
}

/* ARGSUSED */
int samesubjselect (index2, msg, string, pat)
struct index2 *index2;
long msg;
register char *string;
char *pat;
{
	register char *s = field (index2, msg, SUBJECTFIELD);

        for (;;) {
                while (isspace (*s & 0xff))
                        s++;
                if ((*s == 'r' || *s == 'R') && (s[1] == 'e' || s[1] == 'E') &&
                    (s[2] == ';' || s[2] == ':' || isspace (s[2] & 0xff)))
                        s += 3;
                else
                        break;
        }

	return (strcmp (s, string) == 0);
}
