/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:49:04 1991 by burrows        */
/* routines to handle an avl tree */

#include "avlcustom.h"

/* routines are:
 *	avl_add, avl_lookup, avl_iterate, avl_destroy
 */

struct avl {
	AVL_DATA	data;
	struct avl	*left, *right;
	char		balance; /* 0 = even, 1 = left long, 2 = right long */
};

/*
 * Add item to an avl tree.
 *
 * "ptree" is pointer to root of tree.
 * "cmp" is comparison function between data items
 * "data" is the data item to be added
 * "rdata" is the returned data item already in the tree  (may be null)
 * "resolve" resolves conflicts when the same data is added twice (may be null)
 *
 * find position where "data" belongs in tree.
 * if "resolve" is null,
 *    if "data" already in tree, put old data in "*rdata" and return -1
 *    else add "data" to tree
 * else call "resolve" giving "data" and old data as arguments,
 * and put return value into tree.
 *
 * returns
 *	   1 if tree got longer			- success
 *	   0 if it stayed the same length	- success
 * 	   -1 if the item conflicted in the tree - failure
 *			(only happens is "resolve" is non-null)
 *	   -2 if ran out of memory		- failure
 */
extern AVL_DATA *avl_add (        /* struct avl **ptree,
			int (*cmp) (AVL_DATA *a, AVL_DATA *b),
			AVL_DATA *data */);

/* lookup item in an avl tree
 *
 * "ptree" is pointer to root of tree.
 * "cmp" is comparison function between data items
 * "data" is the data item to be deleted
 *
 * Find "data" in tree, and return data item in tree.
 * If it cannot be found, return null.
 *
 * returns
 *	data
 *	null if data not in tree
 */
extern AVL_DATA *avl_lookup (    /* struct avl **ptree,
			int (*cmp) (AVL_DATA *a, AVL_DATA *b),
			AVL_DATA *data */);

/* iterate over an avl tree
 *
 * "ptree" is pointer to root of tree.
 * "func" is a function to be called for each node in the tree
 *
 *   iterate over all nodes in in-order, calling
 *
 *                   (*func) (data, node->data)
 */
extern int avl_iterate (   /* struct avl **ptree,
			int (*func) (char *data, AVL_DATA *a),
			char *data */);

/* delete entire tree */
extern void avl_destroy (/* struct avl **ptree; */);
