/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:47:07 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>
#include <signal.h>

static void precis ();
static int parsedate ();
extern void exit ();
extern void abort ();
extern void perror ();
extern long time ();

/* ARGSUSED */
static void coredump (sig)
int sig;
{
	(void) umask (0);
	(void) mkdir ("/tmp/precisdump", 0777);
	(void) chdir ("/tmp/precisdump");
	abort ();
}

/*ARGSUSED*/
main (argc, argv)
int argc;
char *argv[];
{
	char name[1024];
	unsigned long uid;
	int fd;

#ifdef SETLOCALE
        (void) setlocale (0, "");
#endif

	(void) signal (SIGSEGV, coredump);
	(void) signal (SIGBUS, coredump);

	while (scanf ("%lu %s", &uid, name) == 2) {
		if ((fd = open (name, 0)) < 0) {
			perror (name);
			fd = open ("/dev/null", 0);
		}

		if (fd >= 0) {
			precis (name, fd, uid);
			(void) close (fd);
		}
	}

	if (ferror (stdin) || !feof (stdin)) {
		perror ("precis: stdin");
		exit (2);
	}

	(void) fflush (stdout);
	if ((fsync (fileno (stdout)) < 0 && errno!=EINVAL) || ferror (stdout)) {
		perror ("precis: stdout");
		exit (2);
	}

	exit (0);
}


#define ELEMENTS(x)	(sizeof (x) / sizeof ((x)[0]))

#define SUBJECTFIELD	0
#define FROMFIELD	1
#define TOFIELD		2
#define CCFIELD		3
#define DATEFIELD	4
#define NEWSGROUPSFIELD	5
#define MESSAGEIDFIELD	6
#define INREPLYTOFIELD	7
#define REFERENCESFIELD	8
#define KEYWORDSFIELD	9

static void precis (name, fd, uid)
char *name;
int fd;
unsigned long uid;
{
	static char *headerlines[] = {
		/* code assumes no field has same first letter as another */
		"subject:", "from:", "to:", "cc:", "date:", "newsgroups:",
		"message-id:", "in-reply-to:", "references:", "keywords:"
	};
	static char headerchars[256];
	static char lower[256];
	static int first = 1;
	char buffer[8192];
	char *contents[ELEMENTS (headerlines)];
	int got, nread;
	register int i;
	register char *p, *e, *q;
	char datebuf[256];
	struct stat info;
	long date;
	struct tm *tm;
	int day, month, year;
	int gotspace;

	if (first) {
		for (i = 0; i < ELEMENTS (lower); i++)
			lower[i] = tolower (i);
		for (i = 0; i < ELEMENTS (headerlines); i++) {
			headerchars[headerlines[i][0]] = i+1;
			headerchars[toupper (headerlines[i][0])] = i+1;
		}
		first = 0;
	}

	bzero ((char *) contents, sizeof (contents));

	got = 0;
	while ((nread = read (fd, buffer+got, sizeof (buffer) - got)) > 0)
		got += nread;
	
	if (nread < 0) {
		perror (name);
		return;
	}

	p = buffer;
	e = buffer+got;

	while (e > p && e[-1] != '\n')
		e--;
	
	while (p < e && *p != '\n') {
		i = headerchars[*p & 0xff];
		if (i > 0) {
			i--;
			q = headerlines[i];
			while (*q != '\0' && *q == lower[*p & 0xff])
				q++, p++;
			if (*q == '\0') 		/* matched */
				contents[i] = p;
		}
		while (*p++ != '\n' && *p++ != '\n' &&
		       *p++ != '\n' && *p++ != '\n' &&
		       *p++ != '\n' && *p++ != '\n' &&
		       *p++ != '\n' && *p++ != '\n')
			;
	}

	if ((p = contents[DATEFIELD]) != 0) {
		q = datebuf;
		gotspace = 1;
		while (q < &datebuf[sizeof (datebuf)-2] &&
			p < e && (*p != '\n' || (p+1 < e &&
			isspace (p[1] & 0xff) && p[1] != '\n'))) {
			if (isspace (*p & 0xff)) {
				if (!gotspace)
					*q++ = ' ';
				gotspace = 1;
			} else {
				*q++ = *p;
				gotspace = 0;
			}
			p++;
		}
		*q = '\0';
		if (parsedate (datebuf, &day, &month, &year)) {
			(void) sprintf (datebuf, "%03x%1x%02x",
				year, month, day);
			contents[DATEFIELD] = datebuf;
		}
	}

	if (contents[DATEFIELD] != datebuf) {
		if (fstat (fd, &info) == 0)
			date = info.st_mtime;
		else
			(void) time ((long *) &date);
		tm = localtime (&date);
		(void) sprintf (datebuf, "%03x%1x%02x",
		    1900+tm->tm_year, tm->tm_mon, tm->tm_mday);
		contents[DATEFIELD] = datebuf;
	}


	(void) printf ("%ld", uid);
	for (i = 0; i < ELEMENTS (headerlines); i++) {
		putchar ('\t');
		if ((p = contents[i]) == datebuf) {
			(void) printf ("%s", p);
		} else if (p != 0) {
			int gotspace = 1;

			while (p < e && (*p != '\n' || (p+1 < e &&
				isspace (p[1] & 0xff) && p[1] != '\n'))) {
			    	if (isspace (*p & 0xff)) {
					if (!gotspace)
						putchar (' ');
					gotspace = 1;
				} else {
					putchar (*p);
					gotspace = 0;
				}
				p++;
			}
		}
	}
	putchar ('\n');
}

static int parsedate (s, pday, pmonth, pyear)
char *s;
int *pday;
int *pmonth;
int *pyear;
{
	static char *months[] = {
		"jan", "feb", "mar", "apr", "may", "jun",
		"jul", "aug", "sep", "oct", "nov", "dec"
	};
	static int monthchars[256];
	static char lower[256];
	static int first = 1;
	register char *p, *q, *r;
	register int i, j;
	int day = -1;
	int month = -1;
	int year = -1;

	if (first) {
		for (i = 0; i < ELEMENTS (lower); i++)
			lower[i] = tolower (i);
		for (i = 0; i < ELEMENTS (months); i++) {
			monthchars[months[i][0]] |= 1<<i;
			monthchars[toupper (months[i][0])] |= 1<<i;
		}
		first = 0;
	}

	for (p = s; *p != '\0' && year < 0; ) {
		if (isdigit (*p & 0xff)) {
			if (month >= 0)
				year = atoi (p);
			else
				day = atoi (p);
		} else 
			month = -1;
		if ((j = monthchars[*p & 0xff]) != 0) {
			for (i = 0; i < ELEMENTS (months); i++) {
				if ((j & (1 << i)) != 0) {
					q = months[i];
					r = p;
					while (*q!='\0' && *q==lower[*r & 0xff])
						q++, r++;
					if (*q == '\0') {
						month = i;
						break;
					}
				}
			}
		}

		while (*p != '\0' && !isspace (*p & 0xff))
			p++;
		while (*p != '\0' && isspace (*p & 0xff))
			p++;
	}

	if (year >= 0 && 0 < day && day < 32 && 0 <= month  && month < 12) {
		if (year < 60)
			year += 2000;
		else if (year < 100)
			year += 1900;
		*pday = day;
		*pmonth = month;
		*pyear = year;
		return (1);
	}

	return (0);
}
