/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:47:08 1991 by burrows        */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

extern char *malloc ();
extern void free ();
extern void exit ();
extern void perror ();

static void getnext ();

struct file {
	struct file *next;
	FILE *fp;
	char *name;
	int n;
};

main (argc, argv)
int argc;
char *argv[];
{
	int argn;
	struct file *list = 0;
	register struct file *x, *y;
	register FILE *fp;
	register int c;
	int stdinused = 0;

#ifdef SETLOCALE
        (void) setlocale (0, "");
#endif

	for (argn = 1; argn < argc; argn++) {
		y = (struct file *) malloc (sizeof (struct file));
		if (y == 0) {
			perror ("precismerge: malloc");
			exit (2);
		}
		if (strcmp (argv[argn], "-") == 0 && !stdinused) {
			y->name = "stdin";
			y->fp = stdin;
			stdinused = 1;
		} else if ((y->fp = fopen (argv[argn], "r")) != 0) {
			y->name = argv[argn];
		} else {
			perror (argv[argn]);
			exit (2);
		}
		y->next = list;
		list = y;
		getnext (&list);
		for (x=y; x->next != 0 && y->n > x->next->n; x = x->next)
			;
		if (x != y) {
			list = y->next;
			y->next = x->next;
			x->next = y;
		}
	}

	while (list != 0) {
		(void) printf ("%d\t", list->n);
		fp = list->fp;
		do {
			if ((c = getc (fp)) == EOF) {
				(void) fprintf (stderr,
				    "precismerge: %s: bad format\n",
				    list->name);
				exit (2);
			}
			putchar (c);
		} while (c != '\n');

		getnext (&list);

		if ((y = list) != 0) {
			for (x=y; x->next != 0 && y->n > x->next->n; x=x->next)
				;
			if (x != y) {
				list = y->next;
				y->next = x->next;
				x->next = y;
			}
		}
	}

	(void) fflush (stdout);
	if (ferror (stdout) || (fsync (fileno (stdout)) < 0 && errno!=EINVAL)) {
		perror ("precismerge: stdout");
		exit (2);
	}

	exit (0);
}

static void getnext (pfile)
register struct file **pfile;
{
	register FILE *fp = (*pfile)->fp;
	register int n = 0;
	register int c;

	if ((c = getc (fp)) == EOF) {
		if (ferror (fp)) {
			perror ((*pfile)->name);
			exit (2);
		}
		*pfile = (*pfile)->next;
		return;
	}

	if (isdigit (c)) {
		do  {
			n = n*10 + c-'0';
			c = getc (fp);
		} while (isdigit (c));
	} else
		c = 0;

	if (c != '\t') {
		(void) fprintf (stderr, "precismerge: %s: bad format\n",
			(*pfile)->name);
		exit (2);
	}

	(*pfile)->n = n;
}
