/* Copyright 1991 Digital Equipment Corporation.               */
/* All rights reserved. See copyright notice in attached file COPYRIGHT. */
/* Last modified on Mon Feb 18 15:47:09 1991 by burrows        */
#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>

#ifndef ACTIVEFILE
#define ACTIVEFILE "/usr/lib/news/active"
#endif

extern void perror ();
extern void exit ();
extern void abort ();
extern void free ();
extern char *malloc ();
extern char *realloc ();
extern long atol ();

static char spool[] = "/usr/spool/news/";

static int cmpstr ();
static long freegroup ();
static int cmpmessage ();
static int cmptimes ();
static unsigned long grouptime ();
static unsigned long filetime ();
static unsigned long scan ();
static struct textfile *readtextfile ();
static struct binfile *readbinfile ();
static void addgroup ();
static int writefile ();
static void outname ();
static void finddeletedmessages ();

struct message {
	unsigned long group;
	unsigned long message;
};

static int orignmessagetable;
static int nmessagetable;
static int maxmessagetable;
static struct message *messagetable;
static struct mtime {
	unsigned long mtime;
	unsigned long message;
} *modtimes;

struct group {
	char name[64];
	unsigned long num;
	unsigned long lastmessage;
	unsigned long firstmessage;
};

static int ngrouptable;
static struct group *grouptable;

struct binfile {
	int ndata;
	char *data;
};

struct textfile {
	struct binfile *text;
	int nlines;
	char **lines;
};

main (argc, argv)
int argc;
char *argv[];
{
	struct textfile *active, *use;
	int i, j, len;
	FILE *fp, *dfp;
	char *tmp, *dtmp;
	char name[64];
	unsigned long one, two, newtwo, range;
	unsigned long num;
	unsigned long lastmessage, firstmessage;
	char mode[64];
	struct binfile *msgs;
	int pos;
	unsigned long tstamp;
	char *map;
	char *alreadyseen;
	int all = 0;

	if (argc > 1 && strcmp (argv[1], "-a") == 0) {
		all = 1;
		argc--, argv++;
	}

	if (argc != 7) {
		(void) fprintf (stderr, "usage: scannews oldgroupfile newgroupfile oldmsgfile newmsgfile deletefile timestamp\n");
		exit (2);
	}
	tmp = argv[2];
	dtmp = argv[5];
	tstamp = filetime (argv[6], 0L);

	msgs = readbinfile (argv[3]);
	if (msgs == 0)
		exit (2);
	messagetable = (struct message *) msgs->data;
	nmessagetable = msgs->ndata / sizeof (struct message);
	maxmessagetable = nmessagetable;
	if (msgs->ndata != nmessagetable * sizeof (struct message)) {
		(void) fprintf (stderr, "scannews: %s: bad length\n", argv[3]);
	}
	free ((char *) msgs);
	orignmessagetable = nmessagetable;
	modtimes = 0;

	active = readtextfile (ACTIVEFILE);
	if (active == 0)
		exit (2);

	qsort ((char *) active->lines, active->nlines, sizeof (char *), cmpstr);
	alreadyseen = malloc ((unsigned) active->nlines);
	if (alreadyseen == 0) {
		perror ("malloc");
		exit (2);
	}
	bzero (alreadyseen, active->nlines);

	use = readtextfile (argv[1]);
	if (use == 0)
		exit (2);
	
	fp = fopen (tmp, "w");
	if (fp == NULL) {
		perror (tmp);
		exit (2);
	}

	dfp = fopen (dtmp, "w");
	if (dfp == NULL) {
		perror (dtmp);
		exit (2);
	}

	for (i = 0; i < use->nlines; i++) {
		int count = sscanf (use->lines[i], "%63s %d %d %63s %d %d %d",
			name, &one, &two, mode,
			&num, &lastmessage, &firstmessage);
		
		if (count != 7)
			continue;

		addgroup (name, num, lastmessage, firstmessage, use->nlines);

		j = findinactive (active, name);

		if (j >= 0) {
			alreadyseen[j] = 1;

			len = strlen (active->lines[j]);

			if (strncmp (use->lines[i],active->lines[j],len) != 0 ||
			    use->lines[i][len] != ' ' ||
			    (firstmessage < lastmessage &&
			     tstamp <= grouptime (name, 0L))) {
				two = 0;
				count = sscanf (active->lines[j], "%63s %d %d",
					name, &one, &two);
				if (two < firstmessage)
					two = firstmessage;
				if (two < 1)
					two = 1;
				if (two > one)
					two = one;
				newtwo = two;
				range = one-two+1;
				if ((map = malloc ((unsigned) range))==0) {
					perror ("malloc");
					exit (2);
				}
				bzero (map, (int) range);
				if (scan (num, one, &newtwo, map) != range ||
				    newtwo != firstmessage) {
					finddeletedmessages (dfp, num,
						one, two, map);
				}
				free (map);
				firstmessage = newtwo;
				grouptable[num].firstmessage = newtwo;
			}
			(void) fprintf (fp, "%s %d %d %d\n", active->lines[j],
				    num,
				    grouptable[num].lastmessage,
				    grouptable[num].firstmessage);
		} else 
			(void) fprintf (fp, "%s\n", use->lines[i]);
	}

	for (i = 0; all && i < active->nlines; i++) {
		if (!alreadyseen[i]) {
			int count = sscanf (active->lines[i],
				"%63s %d %d %63s", name, &one, &two, mode);
		
			if (count != 4)
				continue;
			
			num = freegroup ();

			addgroup (name, num, 0L, 1L, use->nlines);

			if (two < 1)
				two = 1;
			if (two > one)
				two = one;
			newtwo = two;
			range = one-two+1;
			if ((map = malloc ((unsigned) range))==0) {
				perror ("malloc");
				exit (2);
			}
			bzero (map, (int) range);
			(void) scan (num, one, &newtwo, map);
			free (map);
			firstmessage = newtwo;
			grouptable[num].firstmessage = newtwo;

			(void) fprintf (stderr, "add group %s\n",
				active->lines[i]);

			(void) fprintf (fp, "%s %d %d %d\n", active->lines[i],
				    num,
				    grouptable[num].lastmessage,
				    grouptable[num].firstmessage);
		}
	}

	pos = orignmessagetable;
	if (nmessagetable > pos+1)
		qsort ((char *) &messagetable[pos], nmessagetable-pos, 
			sizeof (struct message), cmptimes);
	
	while (pos < nmessagetable) {
		messagetable[pos].message =
			modtimes[messagetable[pos].message].message;
		(void) printf ("%d ", pos);
		outname (stdout, grouptable[messagetable[pos].group].name);
		(void) printf ("/%d\n", messagetable[pos].message);
		pos++;
	}

	if (writefile (argv[4], (char *) messagetable,
		nmessagetable * sizeof (struct message)) < 0)
		exit (2);

	(void) fflush (stdout);
	if ((fsync (fileno (stdout)) < 0 && errno!=EINVAL) || ferror (stdout)) {
		perror ("stdout");
		exit (2);
	}

	(void) fflush (fp);
	if ((fsync (fileno (fp)) < 0 && errno != EINVAL) || ferror (fp)) {
		perror (tmp);
		exit (2);
	}
	(void) fclose (fp);

	(void) fflush (dfp);
	if ((fsync (fileno (dfp)) < 0 && errno != EINVAL) || ferror (dfp)) {
		perror (dtmp);
		exit (2);
	}
	(void) fclose (dfp);

	exit (0);
}

static unsigned long grouptime (name, d)
char *name;
unsigned long d;
{
	char buffer[1024];
	register char *p, *q;

	(void) strcpy (buffer, spool);

	q = buffer + strlen (buffer);
	p = name;
	while (*p != 0) {
		if (*p == '.') {
			*q++ = '/';
			p++;
		} else
			*q++ = *p++;
	}
	*q = '\0';

	return (filetime (buffer, d));
}

static long freegroup ()
{
	static long last = 1;

	while (last < ngrouptable) {
		if (grouptable[last].num == 0)
			break;
		last++;
	}
	return (last);
}

static void addgroup (name, num, lastmessage, firstmessage, hint)
char *name;
unsigned long num;
unsigned long lastmessage;
unsigned long firstmessage;
int hint;
{
	int i;

	if (num >= ngrouptable) {
		i = ngrouptable;
		if (num >= hint)
			hint = num + 20;
		ngrouptable = hint;
		grouptable = (struct group *) realloc ((char *) grouptable,
			(unsigned) ngrouptable * sizeof (struct group));
		if (grouptable == 0) {
			perror ("realloc grouptable");
			exit (2);
		}
		bzero ((char *) &grouptable[i],
			(int) ((ngrouptable-i) * sizeof (struct group)));
	}

	if (grouptable[num].num != 0) {
		if (strcmp (name, grouptable[num].name) != 0)
			(void) fprintf (stderr,
			    "scannews: group appears twice: %s\n", name);
		else
			(void) fprintf (stderr,
			    "scannews: group number conflict: %s and %s\n",
				name, grouptable[num].name);
		return;
	}

	grouptable[num].num = num;
	grouptable[num].lastmessage = lastmessage;
	grouptable[num].firstmessage = firstmessage;

	(void) strncpy (grouptable[num].name, name,
		sizeof (grouptable[num].name));
	grouptable[num].name[sizeof (grouptable[num].name) - 1] = '\0';
}

static void addmessage (group, message, hint, t)
unsigned long group;
unsigned long message;
int hint;
unsigned long t;
{
	register int i;

	if (nmessagetable >= maxmessagetable) {
		i = maxmessagetable;
		if (nmessagetable >= hint)
			hint = nmessagetable + 1024;
		maxmessagetable = hint;
		messagetable = (struct message *)realloc ((char *) messagetable,
			(unsigned) maxmessagetable * sizeof (struct message));
		if (messagetable == 0) {
			perror ("realloc messagetable");
			exit (2);
		}
		bzero ((char *) &messagetable[i],
			(int) ((maxmessagetable-i) * sizeof (struct message)));
		
		modtimes = (struct mtime *) realloc ((char *) modtimes,
			(unsigned) (maxmessagetable - orignmessagetable) * 
				sizeof (struct mtime));
		if (modtimes == 0) {
			perror ("realloc modtimes");
			exit (2);
		}
		bzero ((char *) &modtimes[i-orignmessagetable],
			(int) ((maxmessagetable-i) * sizeof (struct mtime)));
	}

	i = nmessagetable - orignmessagetable;
	messagetable[nmessagetable].group = group;
	messagetable[nmessagetable].message = i;
	modtimes[i].mtime = t;
	modtimes[i].message = message;
	nmessagetable++;
}

static void outname (fp, name)
register FILE *fp;
register char *name;
{
	while (*name != '\0') {
		if (*name == '.')
			(void) putc ('/', fp);
		else
			(void) putc (*name, fp);
		name++;
	}
}

static unsigned long filetime (file, d)
char *file;
unsigned long d;
{
	struct stat info;

	if (stat (file, &info) == 0)
		return (info.st_mtime);
	return (d);
}

static unsigned long scan (num, max, pmin, map)
unsigned long num;
unsigned long max;
unsigned long *pmin;
char *map;
{
	char buffer[1024];
	char *p, *q;
	DIR *d;
	struct direct *dp;
	long n;
	unsigned long highest;
	unsigned long mtime, dtime;
	int pos = nmessagetable;
	unsigned long min = max;
	unsigned long omin = *pmin;
	int i, j;
	unsigned long count = 0;

	(void) strcpy (buffer, spool);

	q = buffer + strlen (buffer);
	p = grouptable[num].name;
	while (*p != 0) {
		if (*p == '.') {
			*q++ = '/';
			p++;
		} else
			*q++ = *p++;
	}
	*q = '\0';
	
	d = opendir (buffer);
	if (d == 0) {
		perror (buffer);
		return (count);
	}
	dtime = filetime (buffer, 0L);

	*q++ = '/';

	highest = grouptable[num].lastmessage;
	while ((dp = readdir (d)) != 0) {
		if (dp->d_ino == 0)
			continue;
		if (dp->d_name[0] == '.') {
			if (dp->d_name[1] == '\0')
				continue;
			if (dp->d_name[1] == '.' && dp->d_name[2] == '\0')
				continue;
		}

		n = atol (dp->d_name);

		if (n < omin || n > max)
			continue;

		map[n-omin] = 1;
		count++;

		if (n < min)
			min = n;

		if (n > grouptable[num].lastmessage) {
			(void) strcpy (q, dp->d_name);
			mtime = filetime (buffer, dtime);
		
			addmessage  (num, (unsigned long) n, 0, mtime);
			if (n > highest)
				highest = n;
		}
	}
	grouptable[num].lastmessage = highest;

	(void) closedir (d);

	*pmin = min;

	if (nmessagetable > pos+1) {
		qsort ((char *) &messagetable[pos], nmessagetable-pos, 
			sizeof (struct message), cmpmessage);
	
		mtime = 0;
		j = nmessagetable;
		for (i = pos; i < j; i++) {
			if (modtimes[messagetable[i].message].mtime <= mtime)
				modtimes[messagetable[i].message].mtime =
					++mtime;
			else
				mtime = modtimes[messagetable[i].message].mtime;
		}
	}

	return (count);
}

static void finddeletedmessages (dfp, num, highest, lowest, map)
FILE *dfp;
register unsigned long num;
unsigned long highest;
register unsigned long lowest;
register char *map;
{
	register struct message *m = messagetable;
	register int i = orignmessagetable;

	i++;
	while (--i != 0) {
		if (m->group == num) {
			if (m->message < lowest ||
			    m->message > highest || 
			    !map[m->message - lowest])
				(void) fprintf (dfp, "%d\n", m - messagetable);
		}
		m++;
	}
}

static int cmptimes (a, b)
struct message *a;
struct message *b;
{
	return (modtimes[a->message].mtime - modtimes[b->message].mtime);
}

static int cmpmessage (a, b)
struct message *a;
struct message *b;
{
	if (a->group != b->group)
		return (a->group - b->group);
	return (modtimes[a->message].message - modtimes[b->message].message);
}

static int findinactive (active, string)
struct textfile *active;
register char *string;
{
	int n;
	int min, max, mid;
	int cmp;

	n = strlen (string);
	
	min = 0;
	max = active->nlines;
	while (min < max) {
		mid = (min + max) >> 1;

		cmp = strncmp (active->lines[mid], string, n);
		if (cmp == 0 && active->lines[mid][n] != ' ')
			cmp = 1;

		if (cmp < 0)
			min = mid+1;
		else if (cmp > 0)
			max = mid;
		else
			return (mid);
	}

	return (-1);
}

static int cmpstr (a, b)
char **a, **b;
{
	return (strcmp (*a, *b));
}

static void freebinfile (binfile)
struct binfile *binfile;
{
	free (binfile->data);
	free ((char *) binfile);
}

static void freetextfile (textfile)
struct textfile *textfile;
{
	freebinfile (textfile->text);
	free ((char *) textfile->lines);
	free ((char *) textfile);
}

static struct binfile *readbinfile (file)
char *file;
{
	int fd = open (file, 0);
	struct binfile *binfile;
	struct stat info;
	int got;
	int size;
	char *data;
	int nread;

	if (fd < 0) {
		perror (file);
		return (0);
	}

	if (fstat (fd, &info) < 0) {
		perror (file);
		(void) close (fd);
		return (0);
	}
	size = info.st_size + 1024;

	data = malloc ((unsigned) size + 1);

	if (data == 0) {
		perror ("malloc binfile data");
		(void) close (fd);
		return (0);
	}

	got = 0;
	while ((nread = read (fd, data+got, size - got)) > 0) {
		got += nread;
		if (got > size) {
			size = got + 8192;
			data = realloc (data, (unsigned) size + 1);
			if (data == 0) {
				perror ("realloc binfile data");
				(void) close (fd);
				return (0);
			}
		}
	}

	if (nread < 0) {
		perror (file);
		(void) close (fd);
		free (data);
		return (0);
	}

	(void) close (fd);

	binfile = (struct binfile *) malloc (sizeof (struct binfile));
	if (binfile == 0) {
		perror ("malloc binfile struct");
		free (data);
		return (0);
	}

	binfile->data = data;
	binfile->ndata = got;

	return (binfile);
}

static struct textfile *readtextfile (file)
char *file;
{
	struct binfile *binfile = readbinfile (file);
	struct textfile *textfile;
	register int i;
	register char *p, *e;
	register int maxlines;
	register char **lines;

	if (binfile == 0)
		return (0);

	textfile = (struct textfile *) malloc (sizeof (struct textfile));
	if (textfile == 0) {
		perror ("malloc textfile struct");
		freebinfile (binfile);
		return (0);
	}
	textfile->text = binfile;

	maxlines = (binfile->ndata / 26) + 10;
	lines = (char **) malloc ((unsigned) (maxlines * sizeof (char *)));
	if (lines == 0) {
		perror ("malloc textfile lines");
		freebinfile (binfile);
		free ((char *) textfile);
		return (0);
	}

	binfile->data[binfile->ndata] = '\n';
	for (i = 0, p = binfile->data, e = p+binfile->ndata; p < e; ) {
		lines[i++] = p;
		while (*p++ != '\n' &&  *p++ != '\n' &&
		       *p++ != '\n' &&  *p++ != '\n')
			;
		p[-1] = '\0';
		if (i >= maxlines) {
			maxlines += 1024;
			lines = (char **) realloc ((char *) lines,
				(unsigned) maxlines * sizeof (char *));
			if (lines == 0) {
				perror ("realloc textfile lines");
				freebinfile (binfile);
				free ((char *) textfile);
				return (0);
			}
		}
	}
	textfile->nlines = i;
	textfile->lines = lines;

	return (textfile);
}

static int writefile (file, data, ndata)
char *file;
char *data;
int ndata;
{
	int fd = creat (file, 0666);
	int done;
	int written;

	if (fd < 0) {
		perror (file);
		return (-1);
	}

	for (done = 0; done < ndata; done += written) {
		written = write (fd, data+done, ndata-done);
		if (written < 0) {
			perror (file);
			(void) close (fd);
			return (-1);
		}
	}

	if ((fsync (fd) < 0 && errno != EINVAL)) {
		perror (file);
		(void) close (fd);
		return (-1);
	}
	(void) close (fd);

	return (0);
}
